/****************************************************************************
  $Id: TLAgentApplication.java,v 1.3 1998/12/16 16:28:32 toma Exp $
****************************************************************************/


package webwisdom.tango;


import java.net.*;
import java.io.*;

import webwisdom.tango.consts.Const;
import webwisdom.tango.consts.CAConsts;
import webwisdom.tango.messages.AppEventMessage;
import webwisdom.tango.messages.CAppAnswerEventMessage;


class TLAgentApplicationReceiver extends Thread
{
    private static final String CL="TLAgentApplicationReceiver";


    private DataInputStream inStream;
    private TLAgentApplication agent;


    TLAgentApplicationReceiver(DataInputStream is,
			       TLAgentApplication a)
    {
      inStream=is;
      agent=a;
    }// TLAgentApplicationReceiver _________________________________________


    public void run()
    {
      try
      {
	do
	{
	  int type=inStream.readInt();
	  int packetLen=inStream.readInt();
	  byte[] packetData=new byte[packetLen];
	  inStream.readFully(packetData);
	  AppEventMessage m=new AppEventMessage(type,packetData);
	  agent.receive(m);
	  //System.out.println(CL+".< "+new String(packetData,0));
	}
	while(true);
      }
      catch(EOFException e)
      {
	//System.err.println(CL+": Server disconnected!");
	agent.exit();
      }
      catch(IOException e)
      {
	System.err.println(CL+": "+e);
      }
    }// run ________________________________________________________________


    public String toString()
    {
	return CL+"[agent="+agent+"]";
    }// toString ___________________________________________________________


}// TLAgentApplicationReceiver /////////////////////////////////////////////




/**Tango system proxy for applications.
   This class may be used to create TAgent.
   *@see TAgent*/
public final class TLAgentApplication implements TLAgent
{
    private static final String CL="TLAgentApplication";


    private Socket socket;
    private DataOutputStream outStream;
    private DataInputStream inStream;

    private TLAgentApplicationReceiver receiverThread;

    private TLListener lowlList;


    /**Registers an application in the Tango system.
     *@param args original arguments passed to the program
     *@exception TangoException if cannot connect to Tango*/
    public TLAgentApplication(String[] args)
      throws TangoException
    {
      try
      {
	register(args);
      }
      catch(IOException e)
      {
	throw new TangoException("cannot register: "+e.getMessage());
      }
    }// TLAgentApplication _________________________________________________


    public TLAgentApplication(int AID,int SID,int AT,int port)
      throws TangoException
    {
      try
      {
	register(AID,SID,AT,port);
      }
      catch(IOException e)
      {
	throw new TangoException("cannot register: "+e.getMessage());
      }
    }// TLAgentApplication _________________________________________________


    private void register(int AID,int SID,int AT,int port)
      throws TangoException,IOException
    {
      //System.out.println(CL+".register("+AID+","+SID+","+AT+","+port+")");
      try
      {
	InetAddress host=InetAddress.getLocalHost();
	socket=new Socket(host,port); 
	outStream=new DataOutputStream(socket.getOutputStream());
	inStream=new DataInputStream(socket.getInputStream());
	/*System.out.println(CL+": connected to: "+socketToString(socket)+
	  " (localPort="+socket.getLocalPort()+")");*/
      }
      catch(IOException e)
      {
	throw new TangoException("port "+port+" unreachable");
      }

      receiverThread=new TLAgentApplicationReceiver(inStream,this);
      receiverThread.start();
      (Thread.currentThread()).setPriority(Thread.MIN_PRIORITY);


      //System.out.println(CL+".register1("+AID+","+SID+","+AT+","+port+")");
      outStream.writeInt(AID);
      outStream.writeInt(SID);
      outStream.writeInt(AT);
      outStream.flush();
      //System.out.println(CL+".register2("+AID+","+SID+","+AT+","+port+")");
    }// register ___________________________________________________________


    private void register(String[] args)
      throws TangoException,IOException
    {
      if((args==null)||(args.length<5))
	throw new TangoException("missing arguments: "+args);


      int AID=0,SID,AT;
      int port;

      try
      {
	AID=Integer.valueOf(args[1]).intValue();
	SID=Integer.valueOf(args[2]).intValue();
	AT=Integer.valueOf(args[3]).intValue();
	port=Integer.valueOf(args[4]).intValue();

// 	for(int i=0;i<args.length;)
// 	{
// 	  String pn=args[i++],p=args[i++];
// 	  if(pn.equals("-AID"))
// 	  {
// 	    AID=Integer.valueOf(p).intValue();
// 	  }
// 	  else if(pn.equals("-SID"))
// 	  {
// 	    SID=Integer.valueOf(p).intValue();
// 	  }
// 	  else if(pn.equals("-AT"))
// 	  {
// 	    AT=Integer.valueOf(p).intValue();
// 	  }
// 	  else if(pn.equals("-port"))
// 	  {
// 	    port=Integer.valueOf(p).intValue();
// 	  }
// 	}
      }
      catch(NumberFormatException e)
      {
	throw new TangoException("wrong arguments: "+e.getMessage());
      }
      register(AID,SID,AT,port);
    }// register __________________________________________________________


    /*interface for TLAgentApplicationReceiver*/
    void receive(AppEventMessage m)
    {     
      //System.out.println(CL+".receive("+m+")");
      if(m.getType()==Const.DATA)
      {
	//System.out.println(CL+".receive("+m+"): DATA");
	if(lowlList!=null)
	  lowlList.receive(m);	
      }
      else if(m.getType()==Const.CONTROL)
      {
	//System.out.println(CL+".receive("+m+"): CONTROL");
	AppEventMessage mm;
	byte[] b=m.getData();
	byte t=b[0];
	if((t==CAConsts.ISMASTER)||
	   (t==CAConsts.MASTERNAME)||
	   (t==CAConsts.USERNAME)||
	   (t==CAConsts.HOSTNAME)||
	   (t==CAConsts.DATAPASS)||
	   (t==CAConsts.PARTICIPANTS)||
	   (t==CAConsts.ACTIVEUSERS)||
	   (t==CAConsts.LOCALAPPS))
	  mm=new CAppAnswerEventMessage(b);
	else
	  mm=m;
	//System.out.println(CL+".receive("+m+"): mm="+mm);
	if(lowlList!=null)
	  lowlList.receive(mm);
      }
      else if(m.getType()==Const.PING)
      {
	send(m);
      }
    }// receive ___________________________________________________________


    public void send(AppEventMessage m)
    {
      //System.err.println(CL+".send("+m+")");
      try
      {
	m.send(outStream);
      }
      catch(IOException e)
      {
	System.err.println(CL+".send("+m+"): "+e);
      }
    }// send ______________________________________________________________


    /**This function is not implemented*/
    public void send(int userlist[],AppEventMessage m)
    {
      send(m);
    }// send ______________________________________________________________


    public void add(TLListener tl)
    {
      lowlList=tl;
    }// add _______________________________________________________________


    public void exit()
    {
      System.exit(0);
    }// exit _______________________________________________________________


    public String toString()
    {
      String host=socket.getInetAddress().getHostName();
      int port=socket.getPort();
      return CL+"["+host+":"+port+"]";
    }// toString ___________________________________________________________


}// TLAgentApplication /////////////////////////////////////////////////////


